@@ -63,22 +63,14 @@
 #include "cdo_vlist.h"
 #include "process_int.h"
 #include "cdo_task.h"
-#include "timer.h"
 #include "datetime.h"
 #include "printinfo.h"
 #include "util_date.h"
 #include "param_conversion.h"
+#include "field_functions.h"
 
 #define USE_CDI_STREAM 1
 
-enum
-{
-  HOUR_LEN = 4,
-  DAY_LEN = 6,
-  MON_LEN = 8,
-  YEAR_LEN = 10
-};
-
 struct ReadArguments
 {
   int tsIDnext;
@@ -102,12 +94,10 @@ cdoReadTimestep(void *rarg)
   ReadArguments *readarg = (ReadArguments *) rarg;
   auto &input_vars = readarg->vars;
   RecordInfo *recList = readarg->recList;
-  const auto streamID = readarg->streamID;
+  auto streamID = readarg->streamID;
   int tsIDnext = readarg->tsIDnext;
   int nrecs = readarg->nrecs;
 
-  // timer_start(timer_read);
-
   for (int recID = 0; recID < nrecs; ++recID)
     {
       int varID, levelID;
@@ -139,8 +129,6 @@ cdoReadTimestep(void *rarg)
       input_vars[varID][levelID].nmiss = nmiss;
     }
 
-    // timer_stop(timer_read);
-
 #ifdef USE_CDI_STREAM
   num_recs = streamInqTimestep(streamID, tsIDnext);
 #else
@@ -151,13 +139,13 @@ cdoReadTimestep(void *rarg)
 }
 
 static void
-vlistSetFrequency(int vlistID, int comparelen)
+vlistSetFrequency(int vlistID, int compareDate)
 {
   const char *freq = nullptr;
   // clang-format off
-  if      (comparelen == DAY_LEN)  freq = "day";
-  else if (comparelen == MON_LEN)  freq = "mon";
-  else if (comparelen == YEAR_LEN) freq = "year";
+  if      (compareDate == CMP_DAY)   freq = "day";
+  else if (compareDate == CMP_MONTH) freq = "mon";
+  else if (compareDate == CMP_YEAR)  freq = "year";
   // clang-format on
   if (freq) cdiDefAttTxt(vlistID, CDI_GLOBAL, "frequency", (int) strlen(freq), freq);
 }
@@ -166,60 +154,59 @@ static void
 addOperators(void)
 {
   // clang-format off
-  cdo_operator_add("xtimmin",    FieldFunc_Min,   DATE_LEN, nullptr);
-  cdo_operator_add("xtimmax",    FieldFunc_Max,   DATE_LEN, nullptr);
-  cdo_operator_add("xtimsum",    FieldFunc_Sum,   DATE_LEN, nullptr);
-  cdo_operator_add("xtimmean",   FieldFunc_Mean,  DATE_LEN, nullptr);
-  cdo_operator_add("xtimavg",    FieldFunc_Avg,   DATE_LEN, nullptr);
-  cdo_operator_add("xtimvar",    FieldFunc_Var,   DATE_LEN, nullptr);
-  cdo_operator_add("xtimvar1",   FieldFunc_Var1,  DATE_LEN, nullptr);
-  cdo_operator_add("xtimstd",    FieldFunc_Std,   DATE_LEN, nullptr);
-  cdo_operator_add("xtimstd1",   FieldFunc_Std1,  DATE_LEN, nullptr);
-  cdo_operator_add("xyearmin",   FieldFunc_Min,   YEAR_LEN, nullptr);
-  cdo_operator_add("xyearmax",   FieldFunc_Max,   YEAR_LEN, nullptr);
-  cdo_operator_add("xyearsum",   FieldFunc_Sum,   YEAR_LEN, nullptr);
-  cdo_operator_add("xyearmean",  FieldFunc_Mean,  YEAR_LEN, nullptr);
-  cdo_operator_add("xyearavg",   FieldFunc_Avg,   YEAR_LEN, nullptr);
-  cdo_operator_add("xyearvar",   FieldFunc_Var,   YEAR_LEN, nullptr);
-  cdo_operator_add("xyearvar1",  FieldFunc_Var1,  YEAR_LEN, nullptr);
-  cdo_operator_add("xyearstd",   FieldFunc_Std,   YEAR_LEN, nullptr);
-  cdo_operator_add("xyearstd1",  FieldFunc_Std1,  YEAR_LEN, nullptr);
-  cdo_operator_add("xmonmin",    FieldFunc_Min,   MON_LEN, nullptr);
-  cdo_operator_add("xmonmax",    FieldFunc_Max,   MON_LEN, nullptr);
-  cdo_operator_add("xmonsum",    FieldFunc_Sum,   MON_LEN, nullptr);
-  cdo_operator_add("xmonmean",   FieldFunc_Mean,  MON_LEN, nullptr);
-  cdo_operator_add("xmonavg",    FieldFunc_Avg,   MON_LEN, nullptr);
-  cdo_operator_add("xmonvar",    FieldFunc_Var,   MON_LEN, nullptr);
-  cdo_operator_add("xmonvar1",   FieldFunc_Var1,  MON_LEN, nullptr);
-  cdo_operator_add("xmonstd",    FieldFunc_Std,   MON_LEN, nullptr);
-  cdo_operator_add("xmonstd1",   FieldFunc_Std1,  MON_LEN, nullptr);
+  cdo_operator_add("xtimmin",    FieldFunc_Min,   CMP_DATE, nullptr);
+  cdo_operator_add("xtimmax",    FieldFunc_Max,   CMP_DATE, nullptr);
+  cdo_operator_add("xtimsum",    FieldFunc_Sum,   CMP_DATE, nullptr);
+  cdo_operator_add("xtimmean",   FieldFunc_Mean,  CMP_DATE, nullptr);
+  cdo_operator_add("xtimavg",    FieldFunc_Avg,   CMP_DATE, nullptr);
+  cdo_operator_add("xtimvar",    FieldFunc_Var,   CMP_DATE, nullptr);
+  cdo_operator_add("xtimvar1",   FieldFunc_Var1,  CMP_DATE, nullptr);
+  cdo_operator_add("xtimstd",    FieldFunc_Std,   CMP_DATE, nullptr);
+  cdo_operator_add("xtimstd1",   FieldFunc_Std1,  CMP_DATE, nullptr);
+  cdo_operator_add("xyearmin",   FieldFunc_Min,   CMP_YEAR, nullptr);
+  cdo_operator_add("xyearmax",   FieldFunc_Max,   CMP_YEAR, nullptr);
+  cdo_operator_add("xyearsum",   FieldFunc_Sum,   CMP_YEAR, nullptr);
+  cdo_operator_add("xyearmean",  FieldFunc_Mean,  CMP_YEAR, nullptr);
+  cdo_operator_add("xyearavg",   FieldFunc_Avg,   CMP_YEAR, nullptr);
+  cdo_operator_add("xyearvar",   FieldFunc_Var,   CMP_YEAR, nullptr);
+  cdo_operator_add("xyearvar1",  FieldFunc_Var1,  CMP_YEAR, nullptr);
+  cdo_operator_add("xyearstd",   FieldFunc_Std,   CMP_YEAR, nullptr);
+  cdo_operator_add("xyearstd1",  FieldFunc_Std1,  CMP_YEAR, nullptr);
+  cdo_operator_add("xmonmin",    FieldFunc_Min,   CMP_MONTH, nullptr);
+  cdo_operator_add("xmonmax",    FieldFunc_Max,   CMP_MONTH, nullptr);
+  cdo_operator_add("xmonsum",    FieldFunc_Sum,   CMP_MONTH, nullptr);
+  cdo_operator_add("xmonmean",   FieldFunc_Mean,  CMP_MONTH, nullptr);
+  cdo_operator_add("xmonavg",    FieldFunc_Avg,   CMP_MONTH, nullptr);
+  cdo_operator_add("xmonvar",    FieldFunc_Var,   CMP_MONTH, nullptr);
+  cdo_operator_add("xmonvar1",   FieldFunc_Var1,  CMP_MONTH, nullptr);
+  cdo_operator_add("xmonstd",    FieldFunc_Std,   CMP_MONTH, nullptr);
+  cdo_operator_add("xmonstd1",   FieldFunc_Std1,  CMP_MONTH, nullptr);
   // clang-format on
 }
 
 void *
 XTimstat(void *process)
 {
-  TimeStat timestat_date = TimeStat::MEAN;
-  int64_t vdate = 0, vdate0 = 0;
-  int vtime = 0, vtime0 = 0;
+  constexpr auto timestatDate{ TimeStat::MEAN };
+  CdiDateTime vDateTime0{};
+  CdiDateTime vDateTimeN{};
   CdoStreamID streamID3 = CDO_STREAM_UNDEF;
   int vlistID3, taxisID3 = -1;
   bool lvfrac = false;
-  char indate1[DATE_LEN + 1], indate2[DATE_LEN + 1];
   double vfrac = 1;
 
   cdo_initialize(process);
 
   addOperators();
 
-  const auto operatorID = cdo_operator_id();
+  auto operatorID = cdo_operator_id();
   auto operfunc = cdo_operator_f1(operatorID);
-  const auto comparelen = cdo_operator_f2(operatorID);
+  auto compareDate = cdo_operator_f2(operatorID);
 
-  const auto lmean = (operfunc == FieldFunc_Mean || operfunc == FieldFunc_Avg);
-  const auto lstd = (operfunc == FieldFunc_Std || operfunc == FieldFunc_Std1);
+  auto lmean = (operfunc == FieldFunc_Mean || operfunc == FieldFunc_Avg);
+  auto lstd = (operfunc == FieldFunc_Std || operfunc == FieldFunc_Std1);
   auto lvarstd = (lstd || operfunc == FieldFunc_Var || operfunc == FieldFunc_Var1);
-  const auto lvars2 = lvarstd;
+  auto lvars2 = lvarstd;
   int divisor = (operfunc == FieldFunc_Std1 || operfunc == FieldFunc_Var1);
 
   auto field2_stdvar_func = lstd ? field2_std : field2_var;
@@ -227,7 +214,7 @@ XTimstat(void *process)
 
   if (operfunc == FieldFunc_Mean)
     {
-      const auto oargc = cdo_operator_argc();
+      auto oargc = cdo_operator_argc();
       if (oargc == 1)
         {
           lvfrac = true;
@@ -238,43 +225,38 @@ XTimstat(void *process)
       else if (oargc > 1)
         cdo_abort("Too many arguments!");
     }
-  else
-    {
-      operator_check_argc(0);
-    }
-
-  const int cmplen = DATE_LEN - comparelen;
+  else { operator_check_argc(0); }
 
 #ifdef USE_CDI_STREAM
-  const auto streamID1 = streamOpenRead(cdo_get_stream_name(0));
+  auto streamID1 = streamOpenRead(cdo_get_stream_name(0));
   auto vlistID1 = streamInqVlist(streamID1);
 #else
-  const auto streamID1 = cdo_open_read(0);
+  auto streamID1 = cdo_open_read(0);
   auto vlistID1 = cdo_stream_inq_vlist(streamID1);
 #endif
 
-  const auto vlistID2 = vlistDuplicate(vlistID1);
+  auto vlistID2 = vlistDuplicate(vlistID1);
   vlist_define_timestep_type(vlistID2, operfunc);
 
-  if (cmplen == 0) vlistDefNtsteps(vlistID2, 1);
+  vlistDefNtsteps(vlistID2, (compareDate == CMP_DATE) ? 1 : -1);
 
-  const auto taxisID1 = vlistInqTaxis(vlistID1);
-  const auto taxisID2 = taxisDuplicate(taxisID1);
+  auto taxisID1 = vlistInqTaxis(vlistID1);
+  auto taxisID2 = taxisDuplicate(taxisID1);
   taxisWithBounds(taxisID2);
   if (taxisInqType(taxisID2) == TAXIS_FORECAST) taxisDefType(taxisID2, TAXIS_RELATIVE);
   vlistDefTaxis(vlistID2, taxisID2);
 
-  const auto nvars = vlistNvars(vlistID1);
+  auto nvars = vlistNvars(vlistID1);
 
-  vlistSetFrequency(vlistID2, comparelen);
+  vlistSetFrequency(vlistID2, compareDate);
 
-  const auto streamID2 = cdo_open_write(1);
+  auto streamID2 = cdo_open_write(1);
   cdo_def_vlist(streamID2, vlistID2);
 
   if (Options::cdoDiag)
     {
       char filename[8192];
-      strcpy(filename, cdo_operator_name(operatorID));
+      std::strcpy(filename, cdo_operator_name(operatorID));
       strcat(filename, "_");
       strcat(filename, cdo_get_stream_name(1));
       streamID3 = cdo_open_write(filename);
@@ -286,8 +268,8 @@ XTimstat(void *process)
           vlistDefVarDatatype(vlistID3, varID, CDI_DATATYPE_INT32);
           vlistDefVarMissval(vlistID3, varID, -1);
           cdiDefKeyString(vlistID3, varID, CDI_KEY_UNITS, "");
-          vlistDefVarAddoffset(vlistID3, varID, 0);
-          vlistDefVarScalefactor(vlistID3, varID, 1);
+          cdiDeleteKey(vlistID3, varID, CDI_KEY_ADDOFFSET);
+          cdiDeleteKey(vlistID3, varID, CDI_KEY_SCALEFACTOR);
         }
 
       taxisID3 = taxisDuplicate(taxisID1);
@@ -298,7 +280,7 @@ XTimstat(void *process)
     }
 
   DateTimeList dtlist;
-  dtlist.set_stat(timestat_date);
+  dtlist.set_stat(timestatDate);
   dtlist.set_calendar(taxisInqCalendar(taxisID1));
 
   int FIELD_MEMTYPE = 0;
@@ -338,17 +320,19 @@ XTimstat(void *process)
   tsID++;
   while (true)
     {
-      int nsets = 0;
+      int numSets = 0;
       while (nrecs > 0)
         {
-          dtlist.taxis_inq_timestep(taxisID1, nsets);
-          vdate = dtlist.get_vdate(nsets);
-          vtime = dtlist.get_vtime(nsets);
+          dtlist.taxis_inq_timestep(taxisID1, numSets);
+          auto vDateTime = dtlist.get_vDateTime(numSets);
 
-          if (nsets == 0) SET_DATE(indate2, vdate, vtime);
-          SET_DATE(indate1, vdate, vtime);
+          if (numSets == 0) vDateTime0 = vDateTime;
 
-          if (DATE_IS_NEQ(indate1, indate2, cmplen)) break;
+          if (date_is_neq(vDateTime, vDateTime0, compareDate))
+            {
+              cdo_add_steps(-1);
+              break;
+            }
 
           readarg.tsIDnext = tsID;
           readarg.nrecs = nrecs;
@@ -360,10 +344,7 @@ XTimstat(void *process)
               ltsfirst = false;
               readresult = cdoReadTimestep(&readarg);
             }
-          else
-            {
-              readresult = read_task->wait();
-            }
+          else { readresult = read_task->wait(); }
 
           nrecs = *(int *) readresult;
 
@@ -376,30 +357,30 @@ XTimstat(void *process)
               read_task->start(cdoReadTimestep, &readarg);
             }
 
-          if (nsets == 0)
+          if (numSets == 0)
             {
 #ifdef _OPENMP
 #pragma omp parallel for default(shared) if (maxrecs > 1)
 #endif
-              for (int recID = 0; recID < maxrecs; recID++)
+              for (int recID = 0; recID < maxrecs; ++recID)
                 {
-                  const auto varID = recList[recID].varID;
-                  const auto levelID = recList[recID].levelID;
+                  auto varID = recList[recID].varID;
+                  auto levelID = recList[recID].levelID;
 
                   auto &rsamp1 = samp1[varID][levelID];
                   auto &rvars1 = vars1[varID][levelID];
                   auto &rinput_var = input_vars[curFirst][varID][levelID];
 
-                  const auto nmiss = rinput_var.nmiss;
+                  auto nmiss = rinput_var.nmiss;
 
                   field_copy(rinput_var, rvars1);
                   rvars1.nmiss = nmiss;
                   if (nmiss || !rsamp1.empty())
                     {
-                      const auto fieldsize = rvars1.size;
+                      auto fieldsize = rvars1.size;
                       if (rsamp1.empty()) rsamp1.resize(fieldsize);
 
-                      for (size_t i = 0; i < fieldsize; i++) rsamp1.vec_d[i] = !DBL_IS_EQUAL(rvars1.vec_d[i], rvars1.missval);
+                      for (size_t i = 0; i < fieldsize; ++i) rsamp1.vec_d[i] = !DBL_IS_EQUAL(rvars1.vec_d[i], rvars1.missval);
                     }
                 }
             }
@@ -408,23 +389,23 @@ XTimstat(void *process)
 #ifdef _OPENMP
 #pragma omp parallel for default(shared) if (maxrecs > 1)
 #endif
-              for (int recID = 0; recID < maxrecs; recID++)
+              for (int recID = 0; recID < maxrecs; ++recID)
                 {
-                  const auto varID = recList[recID].varID;
-                  const auto levelID = recList[recID].levelID;
+                  auto varID = recList[recID].varID;
+                  auto levelID = recList[recID].levelID;
 
                   auto &rsamp1 = samp1[varID][levelID];
                   auto &rvars1 = vars1[varID][levelID];
                   auto &rinput_var = input_vars[curFirst][varID][levelID];
 
-                  const auto nmiss = rinput_var.nmiss;
+                  auto nmiss = rinput_var.nmiss;
 
                   if (nmiss || !rsamp1.empty())
                     {
-                      const auto fieldsize = rvars1.size;
-                      if (rsamp1.empty()) rsamp1.resize(fieldsize, nsets);
+                      auto fieldsize = rvars1.size;
+                      if (rsamp1.empty()) rsamp1.resize(fieldsize, numSets);
 
-                      for (size_t i = 0; i < fieldsize; i++)
+                      for (size_t i = 0; i < fieldsize; ++i)
                         if (!DBL_IS_EQUAL(rinput_var.vec_d[i], rvars1.missval)) rsamp1.vec_d[i]++;
                     }
 
@@ -435,49 +416,48 @@ XTimstat(void *process)
                 }
             }
 
-          if (nsets == 0 && lvarstd)
-            for (int recID = 0; recID < maxrecs; recID++)
+          if (numSets == 0 && lvarstd)
+            for (int recID = 0; recID < maxrecs; ++recID)
               {
-                const auto varID = recList[recID].varID;
-                const auto levelID = recList[recID].levelID;
+                auto varID = recList[recID].varID;
+                auto levelID = recList[recID].levelID;
 
                 if (vlistInqVarTimetype(vlistID1, varID) == TIME_CONSTANT) continue;
 
                 field2_moq(vars2[varID][levelID], vars1[varID][levelID]);
               }
 
-          vdate0 = vdate;
-          vtime0 = vtime;
-          nsets++;
+          vDateTimeN = vDateTime;
+          numSets++;
           tsID++;
         }
 
-      if (nrecs == 0 && nsets == 0) break;
+      if (nrecs == 0 && numSets == 0) break;
 
       if (lmean)
         {
 #ifdef _OPENMP
 #pragma omp parallel for default(shared) if (maxrecs > 1)
 #endif
-          for (int recID = 0; recID < maxrecs; recID++)
+          for (int recID = 0; recID < maxrecs; ++recID)
             {
-              const auto varID = recList[recID].varID;
-              const auto levelID = recList[recID].levelID;
+              auto varID = recList[recID].varID;
+              auto levelID = recList[recID].levelID;
 
               if (vlistInqVarTimetype(vlistID1, varID) == TIME_CONSTANT) continue;
 
               if (!samp1[varID][levelID].empty())
                 field2_div(vars1[varID][levelID], samp1[varID][levelID]);
               else
-                fieldc_div(vars1[varID][levelID], (double) nsets);
+                fieldc_div(vars1[varID][levelID], (double) numSets);
             }
         }
       else if (lvarstd)
         {
-          for (int recID = 0; recID < maxrecs; recID++)
+          for (int recID = 0; recID < maxrecs; ++recID)
             {
-              const auto varID = recList[recID].varID;
-              const auto levelID = recList[recID].levelID;
+              auto varID = recList[recID].varID;
+              auto levelID = recList[recID].levelID;
 
               if (vlistInqVarTimetype(vlistID1, varID) == TIME_CONSTANT) continue;
 
@@ -488,30 +468,29 @@ XTimstat(void *process)
               if (!rsamp1.empty())
                 field2_stdvar_func(rvars1, rvars2, rsamp1, divisor);
               else
-                fieldc_stdvar_func(rvars1, rvars2, nsets, divisor);
+                fieldc_stdvar_func(rvars1, rvars2, numSets, divisor);
             }
         }
 
-      if (Options::cdoVerbose)
-        cdo_print("%s %s  vfrac = %g, nsets = %d", date_to_string(vdate0).c_str(), time_to_string(vtime0).c_str(), vfrac, nsets);
+      if (Options::cdoVerbose) cdo_print("%s  vfrac = %g, numSets = %d", datetime_to_string(vDateTimeN), vfrac, numSets);
 
       if (lvfrac && operfunc == FieldFunc_Mean)
-        for (int recID = 0; recID < maxrecs; recID++)
+        for (int recID = 0; recID < maxrecs; ++recID)
           {
-            const auto varID = recList[recID].varID;
-            const auto levelID = recList[recID].levelID;
+            auto varID = recList[recID].varID;
+            auto levelID = recList[recID].levelID;
             auto &rvars1 = vars1[varID][levelID];
 
             if (vlistInqVarTimetype(vlistID1, varID) == TIME_CONSTANT) continue;
 
-            const auto missval = rvars1.missval;
+            auto missval = rvars1.missval;
             if (!samp1[varID][levelID].empty())
               {
-                const auto fieldsize = rvars1.size;
+                auto fieldsize = rvars1.size;
                 size_t irun = 0;
                 for (size_t i = 0; i < fieldsize; ++i)
                   {
-                    if ((samp1[varID][levelID].vec_d[i] / nsets) < vfrac)
+                    if ((samp1[varID][levelID].vec_d[i] / numSets) < vfrac)
                       {
                         rvars1.vec_d[i] = missval;
                         irun++;
@@ -522,19 +501,19 @@ XTimstat(void *process)
               }
           }
 
-      dtlist.stat_taxis_def_timestep(taxisID2, nsets);
+      dtlist.stat_taxis_def_timestep(taxisID2, numSets);
       cdo_def_timestep(streamID2, otsID);
 
       if (Options::cdoDiag)
         {
-          dtlist.stat_taxis_def_timestep(taxisID3, nsets);
+          dtlist.stat_taxis_def_timestep(taxisID3, numSets);
           cdo_def_timestep(streamID3, otsID);
         }
 
-      for (int recID = 0; recID < maxrecs; recID++)
+      for (int recID = 0; recID < maxrecs; ++recID)
         {
-          const auto varID = recList[recID].varID;
-          const auto levelID = recList[recID].levelID;
+          auto varID = recList[recID].varID;
+          auto levelID = recList[recID].levelID;
           auto &rvars1 = vars1[varID][levelID];
 
           if (otsID && vlistInqVarTimetype(vlistID1, varID) == TIME_CONSTANT) continue;
