@@ -54,6 +54,7 @@
 #include <math.h>
 #include <string.h>
 #include <stdint.h>
+#include <float.h>
 
 #include "sphere_part.h"
 #include "geometry.h"
@@ -105,27 +106,26 @@ struct bnd_sphere_part_search {
 };
 
 enum node_type {
-  I_NODE = 0,
-  U_NODE = 1,
-  T_NODE = 2,
+  I_NODE_FULL = 0,
+  I_NODE = 1,
+  U_NODE = 2,
+  T_NODE = 3,
 };
 
 struct temp_partition_data {
+  struct bounding_circle bnd_circle; // has to be first entry
   size_t local_id;
-  struct bounding_circle bnd_circle;
   int node_type;
 };
 
 struct point_id_xyz {
+  double coordinates_xyz[3]; // has to be the first entry
   size_t idx; // index of the point in the coordinates array passed to
               // the constructor
-  double coordinates_xyz[3];
 };
 
 struct point_id_xyz_angle {
-  size_t idx; // index of the point in the coordinates array passed to
-              // the constructor
-  double coordinates_xyz[3];
+  struct point_id_xyz point;
   double cos_angle;
 };
 
@@ -158,6 +158,7 @@ static void init_sphere_part_node(struct sphere_part_node * node) {
    node->I_size = 0;
    node->U_size = 0;
    node->T_size = 0;
+   node->I_angle = SIN_COS_ZERO;
    node->gc_norm_vector[0] = 0;
    node->gc_norm_vector[1] = 0;
    node->gc_norm_vector[2] = 1;
@@ -172,10 +173,114 @@ static struct sphere_part_node * get_sphere_part_node() {
    return node;
 }
 
-static int compare_temp_partition_data(const void * a, const void * b) {
+// computes the norm vector for the plane the partitions the data in half
+// (more or less)
+static inline void compute_gc_norm_vector(
+  void * coords_data, size_t coords_size, size_t coords_count,
+  double prev_gc_norm_vector[], double gc_norm_vector[]) {
 
-  return ((const struct temp_partition_data *)a)->node_type -
-         ((const struct temp_partition_data *)b)->node_type;
+
+   double balance_point[3] = {0.0,0.0,0.0};
+
+   // compute balance point
+   for (size_t i = 0; i < coords_count; ++i) {
+
+     double * coords =
+      (double*)(((unsigned char*)coords_data) + i * coords_size);
+     balance_point[0] += coords[0];
+     balance_point[1] += coords[1];
+     balance_point[2] += coords[2];
+   }
+
+   if ((fabs(balance_point[0]) > 1e-9) ||
+       (fabs(balance_point[1]) > 1e-9) ||
+       (fabs(balance_point[2]) > 1e-9)) {
+      normalise_vector(balance_point);
+   } else {
+     balance_point[0] = prev_gc_norm_vector[2];
+     balance_point[1] = prev_gc_norm_vector[0];
+     balance_point[2] = prev_gc_norm_vector[1];
+   }
+
+   crossproduct_ld(
+      balance_point, prev_gc_norm_vector, gc_norm_vector);
+
+  if ((fabs(gc_norm_vector[0]) > 1e-9) ||
+      (fabs(gc_norm_vector[1]) > 1e-9) ||
+      (fabs(gc_norm_vector[2]) > 1e-9)) {
+     normalise_vector(gc_norm_vector);
+  } else {
+     gc_norm_vector[0] = prev_gc_norm_vector[2];
+     gc_norm_vector[1] = prev_gc_norm_vector[0];
+     gc_norm_vector[2] = prev_gc_norm_vector[1];
+  }
+}
+
+static void swap_partition_data(struct temp_partition_data *a, struct temp_partition_data *b) 
+{ 
+  struct temp_partition_data temp = *a; 
+  *a = *b; 
+  *b = temp; 
+}
+
+static size_t swap_node_type(struct temp_partition_data *part_data, size_t i, int node_type, size_t begin, size_t end)
+{
+  for (size_t j = begin; j < end; ++j)
+    {
+      if (part_data[j].node_type == node_type)
+        {
+          swap_partition_data(&part_data[i], &part_data[j]);
+          return j + 1;
+        }
+    }
+
+  return end + 1;
+}
+
+static void sort_partition_data(struct temp_partition_data *part_data, size_t I_FULL_size, size_t I_size, size_t U_size, size_t T_size)
+{
+   size_t I_begin = I_FULL_size;
+   size_t U_begin = I_FULL_size + I_size;
+   size_t T_begin = I_FULL_size + I_size + U_size;
+
+   size_t I_end = I_begin + I_size;
+   size_t U_end = U_begin + U_size;
+   size_t T_end = T_begin + T_size;
+
+   while (I_end > I_begin && part_data[I_end-1].node_type == I_NODE) I_end--;
+   while (U_end > U_begin && part_data[U_end-1].node_type == U_NODE) U_end--;
+   while (T_end > T_begin && part_data[T_end-1].node_type == T_NODE) T_end--;
+
+   for (size_t i = 0; i < I_FULL_size; ++i)
+     {
+       if (part_data[i].node_type != I_NODE_FULL)
+         {
+           I_begin = swap_node_type(part_data, i, I_NODE_FULL, I_begin, T_end);
+         }
+     }
+
+   I_begin = I_FULL_size;
+   for (size_t i = I_begin; i < I_end; ++i)
+     {
+       if (part_data[i].node_type != I_NODE)
+         {
+           U_begin = swap_node_type(part_data, i, I_NODE, U_begin, U_end);
+           if (U_begin > U_end)
+             {
+               T_begin = swap_node_type(part_data, i, I_NODE, T_begin, T_end);
+             }
+         }
+     }
+
+   U_begin = I_FULL_size + I_size;
+   T_begin = I_FULL_size + I_size + U_size;
+   for (size_t i = U_begin; i < U_end; ++i)
+     {
+       if (part_data[i].node_type != U_NODE)
+         {
+           T_begin = swap_node_type(part_data, i, U_NODE, T_begin, T_end);
+         }
+     }
 }
 
 static void partition_data (
@@ -188,27 +293,14 @@ static void partition_data (
      return;
    }
 
-   double balance_point[3] = {0.0,0.0,0.0};
-
-   // compute balance point
-   for (size_t i = 0; i < num_cell_ids; ++i) {
-
-     balance_point[0] += part_data[i].bnd_circle.base_vector[0];
-     balance_point[1] += part_data[i].bnd_circle.base_vector[1];
-     balance_point[2] += part_data[i].bnd_circle.base_vector[2];
-   }
-
-   normalise_vector(balance_point);
-
-   // compute the great circle that partitions the data in half (more or less)
-
-   crossproduct_ld(
-      balance_point, prev_gc_norm_vector, parent_node->gc_norm_vector);
-   normalise_vector(parent_node->gc_norm_vector);
+   compute_gc_norm_vector(
+     part_data, sizeof(*part_data), num_cell_ids,
+     prev_gc_norm_vector, parent_node->gc_norm_vector);
 
    // partition data into cells that overlap with the great circle and cells
    // that are on side of the circle
 
+   size_t I_FULL_size = 0;
    size_t I_size = 0;
    size_t U_size = 0;
    size_t T_size = 0;
@@ -233,15 +325,27 @@ static void partition_data (
       // if the point intersects with the great circle
       if (compare_angles(diff_angle_gc, curr_bnd_circle.inc_angle) <= 0) {
 
-         // set node type for current cell
-         part_data[i].node_type = I_NODE;
-         I_size++;
+         // if gc_norm_vector or -gc_norm_vector is in the bounding circle
+         if (angle.sin < curr_bnd_circle.inc_angle.sin) {
 
-         struct sin_cos_angle inc_angle =
-            sum_angles_no_check(diff_angle_gc, curr_bnd_circle.inc_angle);
+            // set node type for current cell
+            part_data[i].node_type = I_NODE_FULL;
+            I_FULL_size++;
 
-         if (compare_angles(max_inc_angle, inc_angle) < 0)
-            max_inc_angle = inc_angle;
+            max_inc_angle = SIN_COS_M_PI_2;
+
+         } else {
+
+            // set node type for current cell
+            part_data[i].node_type = I_NODE;
+            I_size++;
+
+            struct sin_cos_angle inc_angle =
+               sum_angles_no_check(diff_angle_gc, curr_bnd_circle.inc_angle);
+
+            if (compare_angles(max_inc_angle, inc_angle) < 0)
+               max_inc_angle = inc_angle;
+         }
 
       // angle > M_PI_2
       } else if (angle.cos < 0.0) {
@@ -258,9 +362,9 @@ static void partition_data (
       }
    }
 
-   qsort(
-      part_data, num_cell_ids, sizeof(*part_data), compare_temp_partition_data);
+   sort_partition_data(part_data, I_FULL_size, I_size, U_size, T_size);
 
+   I_size += I_FULL_size;
    parent_node->I_size = I_size;
    parent_node->U_size = U_size;
    parent_node->T_size = T_size;
@@ -281,36 +385,54 @@ static void partition_data (
          parent_node->I.ivt.head_node = head_node;
          parent_node->I.ivt.num_nodes = I_size;
 
-         for (size_t i = 0; i < I_size; ++i) {
+         // for all bounding circles that include gc_norm_vector or
+         // -gc_norm_vector
+         for (size_t i = 0; i < I_FULL_size; ++i) {
+
+            head_node[i].range.left = -M_PI;
+            head_node[i].range.right = M_PI;
+            head_node[i].value = part_data[i].local_id;
+         }
+
+         for (size_t i = I_FULL_size; i < I_size; ++i) {
 
-            struct sin_cos_angle base_angle, corrected_inc_angle;
-            int big_sum, neg;
             double GCp[3], bVp[3];
             struct bounding_circle curr_bnd_circle = part_data[i].bnd_circle;
 
+            // project the base vector of the current bounding circle onto the
+            // current partitioning great circle
             crossproduct_ld(parent_node->gc_norm_vector,
                             curr_bnd_circle.base_vector, GCp);
             crossproduct_ld(GCp, parent_node->gc_norm_vector, bVp);
+            YAC_ASSERT(
+              (fabs(bVp[0]) > 1e-9) || (fabs(bVp[1]) > 1e-9) ||
+              (fabs(bVp[2]) > 1e-9),
+              "ERROR(partition_data): "
+              "projected vector is nearly identical to gc_norm_vector")
             normalise_vector(bVp);
-            base_angle = get_vector_angle_2(bVp, prev_gc_norm_vector);
-            big_sum =
-              sum_angles(curr_bnd_circle.inc_angle,
-                         get_vector_angle_2(bVp, curr_bnd_circle.base_vector),
-                         &corrected_inc_angle);
 
-            // if the angle is bigger then PI
-            if ((big_sum) || (corrected_inc_angle.sin < 0.0))
-              corrected_inc_angle = SIN_COS_M_PI;
+            // the inc_angle of the bounding circle also needs to be projected
+            // onto the great circle
+            // (the formula for this is:
+            //    new_inc_angle=acos(sin(inc_angle)/cos(base_angle))
+            // To find this formula, you have to determine the longitude of
+            // of longitude circle that has exactly one intersection with the
+            // bounding circle.
+            // (see https://www.dropbox.com/s/82bxzno1mjogbtf/function_2nd_try.pdf)
+            double bnd_circle_lat_cos =
+              get_vector_angle_2(bVp, curr_bnd_circle.base_vector).cos;
+            double inc_angle =
+              M_PI_2 - acos(curr_bnd_circle.inc_angle.sin/bnd_circle_lat_cos);
 
-            struct sin_cos_angle left, right;
-            // base_angle - corrected_inc_angle
-            neg = sub_angles(base_angle, corrected_inc_angle, &left);
-            // base_angle + corrected_inc_angle
-            big_sum = sum_angles(base_angle, corrected_inc_angle, &right);
+            // By definition the previous norm vector is on the current great
+            // great circle. We use this as a reference point for I
 
-            head_node[i].range.left = compute_angle(left) - (neg?2.0*M_PI:0.0);
-            head_node[i].range.right = compute_angle(right) +
-                                       (big_sum?2.0*M_PI:0.0);
+            // compute "distance" of the projected base vector to the reference
+            // point
+            double base_angle = get_vector_angle(bVp, prev_gc_norm_vector);
+
+            head_node[i].range.left = base_angle - inc_angle;
+            head_node[i].range.right = base_angle + inc_angle;
             head_node[i].value = part_data[i].local_id;
          }
 
@@ -373,26 +495,10 @@ static struct point_sphere_part_node * partition_point_data (
   if (curr_tree_depth > *max_tree_depth) *max_tree_depth = curr_tree_depth;
 
   struct point_sphere_part_node * node = xmalloc(1 * sizeof(*node));
-
-  double balance_point[3] = {0.0,0.0,0.0};
-
-  // compute balance point
-
-  for (size_t i = 0; i < num_points; ++i) {
-
-    double * curr_coordinates_xyz = &(points[i].coordinates_xyz[0]);
-    balance_point[0] += curr_coordinates_xyz[0];
-    balance_point[1] += curr_coordinates_xyz[1];
-    balance_point[2] += curr_coordinates_xyz[2];
-  }
-
-  normalise_vector(balance_point);
-
-  // compute the great circle that partitions the data in half (more or less)
-
   double * gc_norm_vector = &(node->gc_norm_vector[0]);
-  crossproduct_ld(balance_point, prev_gc_norm_vector, gc_norm_vector);
-  normalise_vector(gc_norm_vector);
+
+  compute_gc_norm_vector(
+    points, sizeof(*points), num_points, prev_gc_norm_vector, gc_norm_vector);
 
   // angle between a point and the great circle plane
   // acos(dot(gc_norm_vector, point_xyz)) = angle(gc_norm_vector, point_xyz)
@@ -448,7 +554,7 @@ static struct point_sphere_part_node * partition_point_data (
   node->flags = 0;
 
   // check whether the lists are small enough (if not -> partition again)
-  if ((U_size <= threshold) ||(U_size == num_points)) {
+  if ((U_size <= threshold) || (U_size == num_points)) {
 
     node->U = points;
     node->flags |= U_IS_LEAF;
@@ -626,11 +732,6 @@ struct point_sphere_part_search * yac_point_sphere_part_search_mask_new (
   struct point_id_xyz * points =
     get_unique_points(&num_points, coordinates_xyz, ids, mask);
 
-  if (num_points == 0) {
-    free(points);
-    return NULL;
-  }
-
   struct point_sphere_part_search * search = xmalloc(1 * sizeof(*search));
   search->points = xrealloc(points, num_points * sizeof(*points));
 
@@ -657,58 +758,72 @@ static void search_bnd_circle_I_node(
 
   if (node->flags & I_IS_INTERVAL_TREE) {
 
-     struct sin_cos_angle base_angle, corrected_inc_angle;
-     int big_sum, neg;
-     double GCp[3], bVp[3];
-     crossproduct_ld(node->gc_norm_vector,
-                     bnd_circle.base_vector, GCp);
-     crossproduct_ld(GCp, node->gc_norm_vector, bVp);
-     normalise_vector(bVp);
-     base_angle = get_vector_angle_2(bVp, prev_gc_norm_vector);
-     big_sum =
-        sum_angles(
-          bnd_circle.inc_angle, get_vector_angle_2(bVp, bnd_circle.base_vector),
-          &corrected_inc_angle);
+    struct sin_cos_angle angle =
+      get_vector_angle_2(
+        bnd_circle.base_vector, node->gc_norm_vector);
+    angle.cos = fabs(angle.cos);
 
-     // if the angle is bigger then PI
-     if ((big_sum) || (corrected_inc_angle.sin < 0.0))
-       corrected_inc_angle = SIN_COS_M_PI;
-
-     struct sin_cos_angle left, right;
-     // base_angle - corrected_inc_angle
-     neg = sub_angles(base_angle, corrected_inc_angle, &left);
-     // base_angle + corrected_inc_angle
-     big_sum = sum_angles(base_angle, corrected_inc_angle, &right);
-
-     search_interval_tree_buffer->num_overlaps = 0;
-
-     yac_search_interval_tree(
-        node->I.ivt.head_node, node->I.ivt.num_nodes,
-        (struct interval){
-           .left = compute_angle(left) - (neg?2.0*M_PI:0.0),
-           .right = compute_angle(right) + (big_sum?2.0*M_PI:0.0)},
-        search_interval_tree_buffer);
-
-     ENSURE_ARRAY_SIZE(*overlap_cells, *overlap_cells_array_size,
-                       *num_overlap_cells +
-                       search_interval_tree_buffer->num_overlaps);
-
-     for (size_t i = 0; i < search_interval_tree_buffer->num_overlaps;
-          ++i) {
+    // if gc_norm_vector or -gc_norm_vector is in the bounding circle
+    if (compare_angles(angle, bnd_circle.inc_angle) <= 0) {
+      ENSURE_ARRAY_SIZE(*overlap_cells, *overlap_cells_array_size,
+                        *num_overlap_cells + node->I.ivt.num_nodes);
+      for (size_t i = 0; i < node->I.ivt.num_nodes; ++i)
         (*overlap_cells)[(*num_overlap_cells)+i] =
-           node->I.ivt.head_node[
-              search_interval_tree_buffer->overlap_iv[i]].value;
-     }
+          node->I.ivt.head_node[i].value;
+      *num_overlap_cells += node->I.ivt.num_nodes;
+      return;
+    }
 
-     *num_overlap_cells += search_interval_tree_buffer->num_overlaps;
+    double GCp[3], bVp[3];
+
+    // project the base vector of the current bounding circle onto the
+    // current great circle
+    crossproduct_ld(node->gc_norm_vector, bnd_circle.base_vector, GCp);
+    crossproduct_ld(GCp, node->gc_norm_vector, bVp);
+    YAC_ASSERT(
+      (fabs(bVp[0]) > 1e-9) || (fabs(bVp[1]) > 1e-9) || (fabs(bVp[2]) > 1e-9),
+      "ERROR(search_bnd_circle_I_node): "
+      "projected vector is nearly identical to gc_norm_vector")
+    normalise_vector(bVp);
+
+    // the inc_angle of the bounding circle also needs to be projected
+    // onto the great circle (see routine partition data for a detailed
+    // explanation)
+    double bnd_circle_lat_cos =
+      get_vector_angle_2(bVp, bnd_circle.base_vector).cos;
+    double inc_angle =
+      M_PI_2 - acos(bnd_circle.inc_angle.sin/bnd_circle_lat_cos);
+
+    // compute "distance" of the projected base vector to the reference point
+    double base_angle = get_vector_angle(bVp, prev_gc_norm_vector);
+
+    search_interval_tree_buffer->num_overlaps = 0;
+
+    yac_search_interval_tree(
+      node->I.ivt.head_node, node->I.ivt.num_nodes,
+      (struct interval){
+        .left = base_angle - inc_angle,
+        .right = base_angle + inc_angle},
+      search_interval_tree_buffer);
+
+    ENSURE_ARRAY_SIZE(*overlap_cells, *overlap_cells_array_size,
+                      *num_overlap_cells +
+                      search_interval_tree_buffer->num_overlaps);
+
+    for (size_t i = 0; i < search_interval_tree_buffer->num_overlaps; ++i)
+      (*overlap_cells)[(*num_overlap_cells)+i] =
+        node->I.ivt.head_node[
+          search_interval_tree_buffer->overlap_iv[i]].value;
+
+    *num_overlap_cells += search_interval_tree_buffer->num_overlaps;
 
   } else {
 
-     ENSURE_ARRAY_SIZE(*overlap_cells, *overlap_cells_array_size,
-                       *num_overlap_cells + node->I_size);
-     memcpy(*overlap_cells + *num_overlap_cells, node->I.list,
-            node->I_size * sizeof(**overlap_cells));
-     *num_overlap_cells += node->I_size;
+    ENSURE_ARRAY_SIZE(*overlap_cells, *overlap_cells_array_size,
+                      *num_overlap_cells + node->I_size);
+    memcpy(*overlap_cells + *num_overlap_cells, node->I.list,
+    node->I_size * sizeof(**overlap_cells));
+    *num_overlap_cells += node->I_size;
   }
 }
 
@@ -965,7 +1080,7 @@ static void point_search_NN(
           dot_stack[curr_tree_depth] = dot;
           node_stack[curr_tree_depth] = node;
           flags[curr_tree_depth] = 0;
-          skip_U = 0;
+          // skip_U = 0;
           skip_T = 0;
           continue;
         }
@@ -1000,7 +1115,7 @@ static void point_search_NN(
            node_stack[curr_tree_depth] = node;
            flags[curr_tree_depth] = 0;
            skip_U = 0;
-           skip_T = 0;
+           // skip_T = 0;
            continue;
         }
       }
@@ -1073,7 +1188,7 @@ static int point_check_bnd_circle(
    return ret;
 }
 
-static inline int leaf_contains_matching_points(
+static inline int leaf_contains_matching_point(
   struct point_id_xyz * points, size_t num_points, double coordinate_xyz[3],
   size_t ** local_point_ids, size_t * local_point_ids_array_size,
   double (**result_coordinates_xyz)[3],
@@ -1083,57 +1198,19 @@ static inline int leaf_contains_matching_points(
   for (size_t i = 0; i < num_points; ++i) {
 
     // if the points are nearly identical
-    if ((fabs(points[i].coordinates_xyz[0] - coordinate_xyz[0]) < yac_angle_tol) &&
-        (fabs(points[i].coordinates_xyz[1] - coordinate_xyz[1]) < yac_angle_tol) &&
-        (fabs(points[i].coordinates_xyz[2] - coordinate_xyz[2]) < yac_angle_tol)) {
+    if (points_are_identically(points[i].coordinates_xyz, coordinate_xyz)) {
 
       ENSURE_ARRAY_SIZE(*local_point_ids, *local_point_ids_array_size,
-                        total_num_local_point_ids + num_points - i);
-      size_t * local_point_ids_ =
-        (*local_point_ids) + total_num_local_point_ids;
-
-      double (*result_coordinates_xyz_)[3];
-
+                        total_num_local_point_ids + 1);
+      (*local_point_ids)[total_num_local_point_ids] = points[i].idx;
       if (result_coordinates_xyz != NULL) {
         ENSURE_ARRAY_SIZE(*result_coordinates_xyz,
                           *result_coordinates_xyz_array_size,
-                          total_num_local_point_ids + num_points - i);
-        result_coordinates_xyz_ =
-          (*result_coordinates_xyz) + total_num_local_point_ids;
-      } else {
-        result_coordinates_xyz_ = NULL;
+                          total_num_local_point_ids + 1);
+        memcpy((*result_coordinates_xyz) + total_num_local_point_ids,
+               points[i].coordinates_xyz, 3 * sizeof(double));
       }
-
-      local_point_ids_[0] = points[i].idx;
-      if (result_coordinates_xyz_ != NULL) {
-        result_coordinates_xyz_[0][0] = points[i].coordinates_xyz[0];
-        result_coordinates_xyz_[0][1] = points[i].coordinates_xyz[1];
-        result_coordinates_xyz_[0][2] = points[i].coordinates_xyz[2];
-      }
-
-      size_t num_local_point_ids_ = 1;
-
-      for (i += 1; i < num_points; ++i) {
-        // if the points are nearly identical
-        if ((fabs(points[i].coordinates_xyz[0] - coordinate_xyz[0]) < yac_angle_tol) &&
-            (fabs(points[i].coordinates_xyz[1] - coordinate_xyz[1]) < yac_angle_tol) &&
-            (fabs(points[i].coordinates_xyz[2] - coordinate_xyz[2]) < yac_angle_tol)) {
-
-          local_point_ids_[num_local_point_ids_] = points[i].idx;
-          if (result_coordinates_xyz_ != NULL) {
-            result_coordinates_xyz_[num_local_point_ids_][0] =
-              points[i].coordinates_xyz[0];
-            result_coordinates_xyz_[num_local_point_ids_][1] =
-              points[i].coordinates_xyz[1];
-            result_coordinates_xyz_[num_local_point_ids_][2] =
-              points[i].coordinates_xyz[2];
-          }
-          ++num_local_point_ids_;
-        }
-      }
-
-      *num_local_point_ids = num_local_point_ids_;
-
+      *num_local_point_ids = 1;
       return 1;
     }
   }
@@ -1197,13 +1274,13 @@ void yac_point_sphere_part_search_NN(struct point_sphere_part_search * search,
             break;
           } else {
             flags[curr_tree_depth] = T_FLAG;
-            if (curr_node->flags & T_IS_LEAF) {
-              points = (struct point_id_xyz*)(curr_node->T);
-              curr_num_points = curr_node->T_size;
-              break;
-            } else {
-              curr_node = curr_node->T;
-            }
+            YAC_ASSERT(
+              curr_node->flags & T_IS_LEAF,
+              "ERROR(yac_point_sphere_part_search_NN): "
+              "if one branch is empty, the other has to be a leaf");
+            points = (struct point_id_xyz*)(curr_node->T);
+            curr_num_points = curr_node->T_size;
+            break;
           }
         } else curr_node = curr_node->U;
 
@@ -1219,13 +1296,13 @@ void yac_point_sphere_part_search_NN(struct point_sphere_part_search * search,
             break;
           } else {
             flags[curr_tree_depth] = U_FLAG;
-            if (curr_node->flags & U_IS_LEAF) {
-              points = (struct point_id_xyz*)(curr_node->U);
-              curr_num_points = curr_node->U_size;
-              break;
-            } else {
-              curr_node = curr_node->U;
-            }
+            YAC_ASSERT(
+              curr_node->flags & U_IS_LEAF,
+              "ERROR(yac_point_sphere_part_search_NN): "
+              "if one branch is empty, the other has to be a leaf");
+            points = (struct point_id_xyz*)(curr_node->U);
+            curr_num_points = curr_node->U_size;
+            break;
           }
         } else curr_node = curr_node->T;
       }
@@ -1234,7 +1311,7 @@ void yac_point_sphere_part_search_NN(struct point_sphere_part_search * search,
     } while (1);
 
     // if we do not have to do a finer search
-    if (leaf_contains_matching_points(
+    if (leaf_contains_matching_point(
           points, curr_num_points, curr_coordinates_xyz, local_point_ids,
           local_point_ids_array_size, result_coordinates_xyz,
           result_coordinates_xyz_array_size, total_num_local_point_ids,
@@ -1249,9 +1326,10 @@ void yac_point_sphere_part_search_NN(struct point_sphere_part_search * search,
       bnd_circle.base_vector[1] = curr_coordinates_xyz[1];
       bnd_circle.base_vector[2] = curr_coordinates_xyz[2];
       bnd_circle.inc_angle = SIN_COS_M_PI;
+      bnd_circle.sq_crd = DBL_MAX;
 
       check_leaf_NN(
-        points, curr_num_points, curr_coordinates_xyz, &(bnd_circle.inc_angle),
+        points, curr_num_points, curr_coordinates_xyz, &bnd_circle.inc_angle,
         result_coordinates_xyz, result_coordinates_xyz_array_size,
         local_point_ids, local_point_ids_array_size, total_num_local_point_ids,
         num_local_point_ids + i);
@@ -1283,7 +1361,7 @@ static inline int compare_point_id_xyz_angle(const void * a, const void * b) {
 
   if (ret != 0) return ret;
 
-  return (p_a->idx > p_b->idx) - (p_a->idx < p_b->idx);
+  return (p_a->point.idx > p_b->point.idx) - (p_a->point.idx < p_b->point.idx);
 }
 
 static size_t initial_point_bnd_search_NNN(
@@ -1296,9 +1374,19 @@ static size_t initial_point_bnd_search_NNN(
   ENSURE_ARRAY_SIZE(*results, *results_array_size, num_points);
   struct point_id_xyz_angle * results_ = *results;
 
+#ifdef __NEC__
+// vectorization of the following loop leads to failues of
+//
+// test_couple_config_parallel1
+// test_interp_method_nnn_parallel
+// test_interp_method_hcsbb_parallel
+//
+// with NEC compiler when CFLAGS='-O2'
+#pragma _NEC novector
+#endif
   for (size_t i = 0; i < num_points; ++i) {
 
-    *(struct point_id_xyz*)(results_ + i) = points[i];
+    results_[i].point = points[i];
     results_[i].cos_angle = clamp_abs_one(
       points[i].coordinates_xyz[0] * point_coordinates_xyz[0] +
       points[i].coordinates_xyz[1] * point_coordinates_xyz[1] +
@@ -1345,9 +1433,8 @@ static inline struct sin_cos_angle check_leaf_NNN(
     // if the point is worse than the currently best point
     if (curr_cos_angle < min_cos_angle) continue;
 
-    struct point_id_xyz_angle point;
-    *(struct point_id_xyz*)(&point) = points[i];
-    point.cos_angle = curr_cos_angle;
+    struct point_id_xyz_angle point =
+      {.point = points[i], .cos_angle = curr_cos_angle};
 
     // insert point
     size_t j;
@@ -1383,7 +1470,7 @@ static inline struct sin_cos_angle check_leaf_NNN(
 
     return
       get_vector_angle_2(
-        results_[num_results_-1].coordinates_xyz, point_coordinates_xyz);
+        results_[num_results_-1].point.coordinates_xyz, point_coordinates_xyz);
   } else return curr_angle;
 }
 
@@ -1396,7 +1483,8 @@ static void point_search_NNN(
 
   struct sin_cos_angle angle =
     get_vector_angle_2(
-      (*results)[(*num_results)-1].coordinates_xyz, point_coordinates_xyz);
+      (*results)[(*num_results)-1].point.coordinates_xyz,
+                       point_coordinates_xyz);
 
   // if we have already found at least n exactly matching points
   if ((*num_results >= n) && (angle.sin <= yac_angle_tol)) return;
@@ -1432,7 +1520,7 @@ static void point_search_NNN(
           dot_stack[curr_tree_depth] = dot;
           node_stack[curr_tree_depth] = node;
           flags[curr_tree_depth] = 0;
-          skip_U = 0;
+          // skip_U = 0;
           skip_T = 0;
           continue;
         }
@@ -1464,7 +1552,7 @@ static void point_search_NNN(
           node_stack[curr_tree_depth] = node;
           flags[curr_tree_depth] = 0;
           skip_U = 0;
-          skip_T = 0;
+          // skip_T = 0;
           continue;
         }
       }
@@ -1494,6 +1582,8 @@ void yac_point_sphere_part_search_NNN(struct point_sphere_part_search * search,
                                       size_t * local_point_ids_array_size,
                                       size_t * num_local_point_ids) {
 
+  if (num_points == 0) return;
+
   if (cos_angles != NULL)
     ENSURE_ARRAY_SIZE(*cos_angles, *cos_angles_array_size, num_points * n);
 
@@ -1513,7 +1603,7 @@ void yac_point_sphere_part_search_NNN(struct point_sphere_part_search * search,
                         total_num_local_points);
 
       for (size_t i = num_points - 1, offset = total_num_local_points - 1;
-           i < num_points; i--) {
+           i != (size_t)-1; i--) {
 
         for (size_t j = 0; j < num_local_point_ids[i]; ++j, --offset)
           (*cos_angles)[offset] = (*cos_angles)[i];
@@ -1642,12 +1732,12 @@ void yac_point_sphere_part_search_NNN(struct point_sphere_part_search * search,
 
     for (size_t j = 0; j < num_results; ++j) {
 
-      local_point_ids_[j] = results[j].idx;
+      local_point_ids_[j] = results[j].point.idx;
       if (cos_angles_ != NULL) cos_angles_[j] = results[j].cos_angle;
       if (result_coordinates_xyz_ != NULL) {
-        result_coordinates_xyz_[j][0] = results[j].coordinates_xyz[0];
-        result_coordinates_xyz_[j][1] = results[j].coordinates_xyz[1];
-        result_coordinates_xyz_[j][2] = results[j].coordinates_xyz[2];
+        result_coordinates_xyz_[j][0] = results[j].point.coordinates_xyz[0];
+        result_coordinates_xyz_[j][1] = results[j].point.coordinates_xyz[1];
+        result_coordinates_xyz_[j][2] = results[j].point.coordinates_xyz[2];
       }
     }
 
@@ -1723,14 +1813,24 @@ static void search_point(struct sphere_part_node * node,
 
       if (node->flags & I_IS_INTERVAL_TREE) {
 
+         // project the point onto the current great circle
          double GCp[3], bVp[3];
          crossproduct_ld(node->gc_norm_vector, point, GCp);
          crossproduct_ld(GCp, node->gc_norm_vector, bVp);
-         normalise_vector(bVp);
-         double base_angle = get_vector_angle(bVp, prev_gc_norm_vector);
 
-         struct interval search_interval =
-          {.left=base_angle, .right=base_angle};
+         struct interval search_interval;
+         // if the projected point does not coincide with the norm vector of
+         // the plane through the great circle
+         if ((fabs(bVp[0]) > 1e-9) || (fabs(bVp[1]) > 1e-9) ||
+             (fabs(bVp[2]) > 1e-9)) {
+            normalise_vector(bVp);
+            double base_angle = get_vector_angle(bVp, prev_gc_norm_vector);
+            search_interval.left=base_angle;
+            search_interval.right=base_angle;
+         } else {
+            search_interval.left = -M_PI;
+            search_interval.right = M_PI;
+         }
 
          search_interval_tree_buffer->num_overlaps = 0;
 
