@@ -1,18 +1,8 @@
 /*
-  This file is part of CDO. CDO is a collection of Operators to
-  manipulate and analyse Climate model Data.
+  This file is part of CDO. CDO is a collection of Operators to manipulate and analyse Climate model Data.
 
-  Copyright (C) 2003-2021 Uwe Schulzweida, <uwe.schulzweida AT mpimet.mpg.de>
-  See COPYING file for copying and redistribution conditions.
+  Author: Uwe Schulzweida
 
-  This program is free software; you can redistribute it and/or modify
-  it under the terms of the GNU General Public License as published by
-  the Free Software Foundation; version 2 of the License.
-
-  This program is distributed in the hope that it will be useful,
-  but WITHOUT ANY WARRANTY; without even the implied warranty of
-  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-  GNU General Public License for more details.
 */
 #ifndef PMLIST_H
 #define PMLIST_H
@@ -31,33 +21,37 @@ struct KeyValues
   std::vector<std::string> values;
 };
 
-class // KVList
+// clang-format off
+class  // KVList
 #ifdef WARN_UNUSED
 [[gnu::warn_unused]]
 #endif
 KVList : public std::list<KeyValues>
+// clang-format on
 {
 public:
   std::string name;
 
-  void print() const;
-  int parse_arguments(const int argc, const std::vector<std::string> &argv);
+  void print(FILE *fp = stderr) const;
+  int parse_arguments(int argc, const std::vector<std::string> &argv);
   const KeyValues *search(const std::string &key) const;
   void remove(const std::string &inkey);
   void append(const char *, const char *const *, int);
   char *get_first_value(const char *key, const char *replacer);
 };
 
-class // PMList
+// clang-format off
+class  // PMList
 #ifdef WARN_UNUSED
 [[gnu::warn_unused]]
 #endif
 PMList : public std::list<KVList>
+// clang-format on
 {
 public:
   const KVList *searchKVListVentry(const std::string &key, const std::string &value, const std::vector<std::string> &entry);
   const KVList *getKVListVentry(const std::vector<std::string> &entry);
-  void print();
+  void print(FILE *fp = stderr);
   void read_namelist(FILE *fp, const char *name);
   void read_cmor_table(FILE *fp, const char *name);
 };
